import csv
import fractions
import math
import os
from functools import reduce
from itertools import combinations, product
from . import pitch, utilities_general, utilities_music, constants


class PitchCollection():

    def __init__(
        self,
        pc = [(1, 1), (2, 1)],
        rp = "A4",
        rf = 440.0,
        tuneable_intervals = constants.SABAT_SCHWEINITZ_TUNEABLE_INTERVALS,
        precision = 5):
        self.pc_raw = list(dict.fromkeys(pc))
        self.reference_pitch = rp
        self.reference_freq = rf
        self.precision = precision
        self.tuneable_intervals = utilities_general.tuples_to_fractions(tuneable_intervals)
        self.info_by_pitch = self._info_by_pitch(self.pc_raw)
        if isinstance(self.info_by_pitch, list):
            self.sort_by(sort_by="ratios")
            intervals_and_resultant_tones = self._intervals_and_resultant_tones()
            self.intervals = intervals_and_resultant_tones[0]
            self.tuneable_intervals = intervals_and_resultant_tones[1][0]
            self.tuneable_pitch_pairs = intervals_and_resultant_tones[1][1]
            self.non_tuneable_intervals = intervals_and_resultant_tones[1][2]
            self.non_tuneable_pitch_pairs = intervals_and_resultant_tones[1][3]
            self.difference_tones = intervals_and_resultant_tones[2]
            self.tuneable_difference_tones = intervals_and_resultant_tones[3][0]
            self.tuneable_difference_tone_pitch_pairs = intervals_and_resultant_tones[3][1]
            self.non_tuneable_difference_tones = intervals_and_resultant_tones[3][2]
            self.non_tuneable_difference_tone_pitch_pairs = intervals_and_resultant_tones[3][3]
            self.summation_tones = intervals_and_resultant_tones[4]
            self.tuneable_summation_tones = intervals_and_resultant_tones[5][0]
            self.tuneable_summation_tone_pitch_pairs = intervals_and_resultant_tones[5][1]
            self.non_tuneable_summation_tones = intervals_and_resultant_tones[5][2]
            self.non_tuneable_summation_tone_pitch_pairs = intervals_and_resultant_tones[5][3]
            self.pc_plus_resultant_tones = self._pc_plus_resultant_tones()
            self.pc_plus_resultant_tones_as_harmonics = self._harmonics(self.pc_plus_resultant_tones)

            self.inversion = self._inversion()
            self.inversion_harmonics = self._harmonics(self.inversion)
            self.intervals_sequential = self._intervals_sequential()
                
            self.avg_ratio = sum(self.ratios) / len(self.ratios)
            self.avg_freq = sum(self.freqs) / len(self.freqs)
            self.avg_keynum = utilities_music.cpsmidi(self.avg_freq, self.reference_freq, self.reference_keynum)
            self.minimum_ratio = min(self.ratios)
            self.minimum_freq = min(self.freqs)
            self.minimum_keynum = min(self.keynums)
            self.maximum_ratio = max(self.ratios)
            self.maximum_freq = max(self.freqs)
            self.maximum_keynum = max(self.keynums)
            self.ratio_span = self.maximum_ratio / self.minimum_ratio
            self.freq_span = self.maximum_freq - self.minimum_freq
            self.keynum_span = self.maximum_keynum - self.minimum_keynum
            self.cents_span = self.keynum_span * 100
                
            self.harmonics = self._harmonics(self.ratios)
            self.periodicity_pitch = self._periodicity_pitch()
            self.least_common_partial = self._least_common_partial()[0]
            self.least_common_partial_freq = self._least_common_partial()[1]
            self.constituent_primes = sorted(list(dict.fromkeys([p for lop in self.constituent_primes_by_pitch for p in lop])))
            self.hd_sum = self._hd_sum()
            self.hd_avg = self.hd_sum / len(self.pc_raw)
            self.harmonic_intersection = self._harmonic_intersection()
            self.harmonic_disjunction = 1 - self.harmonic_intersection

    def print_info(self, variety = "basic"):
        strings_to_print = self._create_strings_for_print_and_txt(variety = variety)
        for x in strings_to_print:
            print(x)

    def sort_by(self, sort_by = "ratios"):
        parameter_index_pairs = [
            ["ratios", 1], 
            ["keynum classes", 5], 
            ["normalized ratios", 10], 
            ["harmonic distances", 11], 
            ["normalized harmonic distances", 14]]
        if sort_by == "reset":
            self.info_by_pitch = self._info_by_pitch(self.pc_raw)
        else:
            sort_by_index = -1
            for pip in parameter_index_pairs:
                if pip[0] == sort_by:
                    sort_by_index = pip[1]
            self.info_by_pitch.sort(key=lambda x: x[sort_by_index])
            self._info_by_parameter()
            self.inversion = self._inversion()
            self.intervals_sequential = self._intervals_sequential()
            if sort_by_index < 0:
                raise ValueError("cannot sort by this parameter")

    def transpose(self, interval):
        transposition_ratio = pitch.Pitch(p = interval).ratio
        transposed_pitches = [x * transposition_ratio for x in self.ratios]
        self.update(pc = transposed_pitches)

    def update(
        self, 
        pc = False, 
        rp = False, 
        rf = False, 
        tuneable_intervals = False, 
        precision = False):
        if not rp:
            rp = self.reference_pitch
        if not rf:
            rf = self.reference_freq
        if not pc:
            pc = self.pc_raw
        if not tuneable_intervals:
            tuneable_intervals = self.tuneable_intervals
        if not precision:
            precision = self.precision
        self.__init__(
            pc = pc, 
            rp = rp, 
            rf = rf, 
            tuneable_intervals = tuneable_intervals, 
            precision = precision)    

    def write_info_to_csv(self, output_directory = False, filename = "pitch_collection_info.csv"):
        final_info = []
        if output_directory == False:
            output_directory = os.getcwd()
        path_to_write_file = output_directory + "/" + filename
        data_types = [
            "", 
            "harm.", 
            "ratio", 
            "monzo", 
            "freq (Hz)", 
            "keynum", 
            "keynum cl.", 
            "primes", 
            "hd", 
            "HEJI", 
            "12-ED2"]
        count = 0
        processed_info = []
        for x in self.info_by_pitch:
            count += 1
            harm = x[0]
            ratio = x[1]
            monzo = x[2]
            freq = x[3]
            keynum = x[4]
            keynum_class = x[5]
            primes = x[7]
            harmonic_distance = x[13]
            notation = x[8]
            letter_name_and_octave_and_cents = x[9]
            stats_to_be_added = [
                count,
                str(harm),
                "'" + str(ratio) + "'",
                str(monzo),
                utilities_general.convert_data_to_readable_string(freq, precision = self.precision),
                utilities_general.convert_data_to_readable_string(keynum, precision = self.precision),
                utilities_general.convert_data_to_readable_string(keynum_class, precision = self.precision),
                str(primes),
                utilities_general.convert_data_to_readable_string(harmonic_distance, precision = self.precision),
                str(notation),
                "'" + str(letter_name_and_octave_and_cents) + "'"]
            processed_info.append(stats_to_be_added)
        final_info = [data_types] + processed_info
        with open(path_to_write_file, "w") as output:
            writer = csv.writer(output, lineterminator='\n')
            writer.writerows(final_info)
        print("file written to " + path_to_write_file)

    def write_info_to_txt(self, variety = "all", output_directory = False, filename = "pitch_collection_info.txt"):
        strings_to_write = self._create_strings_for_print_and_txt(variety = variety)
        if output_directory == False:
            output_directory = os.getcwd()
        path_to_write_file = output_directory + "/" + filename
        with open(path_to_write_file, "w") as output:
            for x in strings_to_write:
                output.write(x + "\n")
        print("file written to " + path_to_write_file)

    def _create_strings_for_print_and_txt(self, variety = "basic"):
        basic_info_strings = [
            "",
            "BASIC INFO",
            "ratios: " + str([utilities_general.convert_data_to_readable_string(x) for x in self.ratios]),
            "frequencies (Hz): " + str([utilities_general.convert_data_to_readable_string(x, precision = self.precision) for x in self.freqs]),
            "MIDI key numbers: " + str([utilities_general.convert_data_to_readable_string(x, precision = self.precision) for x in self.keynums]),
            "Helmholtz-Ellis notations (text string, letter name): " + str(self.notations),
            "12-ED2 pitch and cent deviations: " + str(self.letter_names_and_octave_and_cents),
            "harmonic constellation: " + self._harmonics_to_proportional_ratio_string(self.harmonics),
            "sequential intervals: " + str([utilities_general.convert_data_to_readable_string(x) for x in self.intervals_sequential]),
            "normalized ratios: " + str([utilities_general.convert_data_to_readable_string(x) for x in self.normalized_ratios]),
            "inversion: " + str([utilities_general.convert_data_to_readable_string(x) for x in self.inversion]),
            ""]
        
        quantitative_info_strings = [
            "",
            "QUANTITATIVE INFO",
            "average ratio: " + str(self.avg_ratio),
            "minimum ratio: " + str(self.minimum_ratio),
            "maximum ratio: " + str(self.maximum_ratio),
            "ratio span: " + str(self.ratio_span),
            "average frequency (Hz): " + utilities_general.convert_data_to_readable_string(self.avg_freq, precision = self.precision),
            "minimum frequency (Hz): " + utilities_general.convert_data_to_readable_string(self.minimum_freq, precision = self.precision),
            "maximum frequency (Hz): " + utilities_general.convert_data_to_readable_string(self.maximum_freq, precision = self.precision),
            "frequency span (Hz): " + utilities_general.convert_data_to_readable_string(self.freq_span, precision = self.precision),
            "average MIDI key number: " + utilities_general.convert_data_to_readable_string(self.avg_keynum, precision = self.precision),
            "minimum MIDI key number: " + utilities_general.convert_data_to_readable_string(self.minimum_keynum, precision = self.precision),
            "maximum MIDI key number: " + utilities_general.convert_data_to_readable_string(self.maximum_keynum, precision = self.precision),
            "MIDI key number span: " + utilities_general.convert_data_to_readable_string(self.keynum_span, precision = self.precision),
            "span in cents: " + utilities_general.convert_data_to_readable_string(self.cents_span, precision = self.precision),
            ""]
        
        analytic_info_strings = [
            "",
            "ANALYTIC INFO",
            "all intervals: " + str([utilities_general.convert_data_to_readable_string(x) for x in self.intervals]),
            "tuneable intervals: " + str([utilities_general.convert_data_to_readable_string(x) for x in self.tuneable_intervals]),
            "periodicity pitch (Hz): " + utilities_general.convert_data_to_readable_string(self.periodicity_pitch, precision = self.precision),
            "least common partial (of periodicity pitch): " + str(self.least_common_partial),
            "least common partial frequency (Hz): " + utilities_general.convert_data_to_readable_string(self.least_common_partial_freq, precision = self.precision),
            "constituent primes: " + str(self.constituent_primes),
            "harmonic distance sum: " + str(round(self.hd_sum, self.precision)),
            "average harmonic distance: " + str(round(self.hd_avg, self.precision)),
            "harmonic intersection: " + str(self.harmonic_intersection) + " (" + str(round(float(self.harmonic_intersection), self.precision)) + ")",
            "harmonic disjunction: " + str(self.harmonic_disjunction) + " (" + str(round(float(self.harmonic_disjunction), self.precision)) + ")",
            ""]

        if variety == "normalized" or variety == "all":
            normalized_ci = PitchCollection(pc = self.normalized_ratios, 
                rp = self.reference_pitch, 
                rf = self.reference_freq,
                precision = self.precision)
            normalized_info_strings = normalized_ci._create_strings_for_print_and_txt(variety = "basic")
            del normalized_info_strings[-3]
            normalized_info_strings[1] = "NORMALIZED INFO"

        if variety == "inversion" or variety == "all":
            inversion_ci = PitchCollection(pc = self.inversion, 
                rp = self.reference_pitch, 
                rf = self.reference_freq,
                precision = self.precision)
            inversion_info_strings = inversion_ci._create_strings_for_print_and_txt(variety = "basic")
            del inversion_info_strings[-2]
            inversion_info_strings[1] = "INVERSION INFO"

        if variety == "resultants" or variety == "all":
            difference_tones_ci = PitchCollection(
                pc = self.difference_tones, 
                rp = self.reference_pitch, 
                rf = self.reference_freq,
                precision = self.precision)
            difference_tones_info_strings = difference_tones_ci._create_strings_for_print_and_txt(variety = "basic")
            difference_tones_info_strings[1] = "FIRST-ORDER DIFFERENCE TONES"
            tuneable_difference_tones_strings = "tuneable ratios: " + str([utilities_general.convert_data_to_readable_string(x) for x in self.tuneable_difference_tones])
            difference_tones_info_strings = difference_tones_info_strings[:3] + [tuneable_difference_tones_strings] + difference_tones_info_strings[3:][:-5] + [""]
            summation_tones_ci = PitchCollection(
                pc = self.summation_tones, 
                rp = self.reference_pitch, 
                rf = self.reference_freq,
                precision = self.precision)
            summation_tones_info_strings = summation_tones_ci._create_strings_for_print_and_txt(variety = "basic")
            summation_tones_info_strings[1] = "FIRST-ORDER SUMMATION TONES"
            tuneable_summation_tones_strings = "tuneable ratios: " + str([utilities_general.convert_data_to_readable_string(x) for x in self.tuneable_summation_tones])
            summation_tones_info_strings = summation_tones_info_strings[:3] + [tuneable_summation_tones_strings] + summation_tones_info_strings[3:][:-5] + [""]
            resultant_tones_info_strings = difference_tones_info_strings + summation_tones_info_strings[1:]
        
        reference_info_strings = pitch.Pitch(
            rp = self.reference_pitch,
            rf = self.reference_freq,
            precision = self.precision).create_strings_for_print_and_txt(variety = "reference")

        if variety == "basic":
            output = basic_info_strings
        elif variety == "quantitative":
            output = quantitative_info_strings
        elif variety == "analytic":
            output = analytic_info_strings
        elif variety == "normalized":
            output = normalized_info_strings
        elif variety == "inversion":
            output = inversion_info_strings
        elif variety == "resultants":
            output = resultant_tones_info_strings
        elif variety == "reference":
            output = reference_info_strings
        elif variety == "all":
            output = basic_info_strings + quantitative_info_strings[1:] + analytic_info_strings[1:] + normalized_info_strings[1:] + inversion_info_strings[1:] + resultant_tones_info_strings[1:] + reference_info_strings[1:]
        return(output)

    def _harmonic_intersection(self):
        harmonics = list(set(self.harmonics))
        harmonics_length = len(harmonics)
        harmonics = [ int(h / reduce(math.gcd, harmonics)) for h in harmonics ]
        output = fractions.Fraction(0, 1)
        for i in range(1, harmonics_length + 1):
            combos = combinations(harmonics, i)
            if i % 2 == 0:
                new_values = [ [-1, utilities_general.lcm(c)] for c in combos ]
            else:
                new_values = [ [1, utilities_general.lcm(c)] for c in combos ]
            for nv in new_values:
                output = output + fractions.Fraction(nv[0], nv[1])
        return(output)

    def _harmonics(self, ratios):
        ratio_multiplier = reduce(fractions.gcd,(ratios)).denominator
        harmonics = [int(x * ratio_multiplier) for x in ratios]
        return(harmonics)

    def _harmonics_to_proportional_ratio_string(self, harmonics):
        output = ""
        for i, h in enumerate(harmonics):
            output = output + str(h)
            if i < (len(harmonics) - 1):
                output = output + ":"
        return(output)

    def _hd_sum(self):
        hd_sum = sum(self.harmonic_distances)
        return(hd_sum)

    def _info_by_parameter(self):
        info_by_parameter = utilities_general.flop(self.info_by_pitch)
        self.ratios = info_by_parameter[1]
        self.monzos = info_by_parameter[2]
        self.freqs = info_by_parameter[3]
        self.keynums = info_by_parameter[4]
        self.keynum_classes = info_by_parameter[5]
        self.distances_from_reference = info_by_parameter[6]
        self.constituent_primes_by_pitch = info_by_parameter[7]
        self.notations = info_by_parameter[8]
        self.letter_names_and_octave_and_cents = info_by_parameter[9]
        self.normalized_ratios = list(dict.fromkeys(sorted(info_by_parameter[10])))
        self.normalized_monzos = info_by_parameter[11]
        self.nums_symbols = info_by_parameter[12]
        self.harmonic_distances = info_by_parameter[13]
        self.normalized_harmonic_distances = info_by_parameter[14]
        return(info_by_parameter)

    def _info_by_pitch(self, pc):
        ratios = []
        pc_processed = []
        output = []
        for p in pc:
            pci = pitch.Pitch(p = p,
                rp = self.reference_pitch, 
                rf = self.reference_freq)
            if isinstance(pci.rk_and_fo, str) == False:
                ratios.append(pci.ratio)
                pc_processed.append([pci.ratio, 
                    pci.monzo, 
                    pci.freq, 
                    pci.keynum, 
                    pci.keynum_class, 
                    pci.distance_in_cents_from_reference,
                    pci.constituent_primes, 
                    pci.notation, 
                    pci.letter_name_and_octave_and_cents, 
                    pci.normalized_ratio, 
                    pci.normalized_monzo, 
                    pci.num_symbols, 
                    pci.harmonic_distance, 
                    pci.normalized_harmonic_distance])
                self.reference_keynum = pci.reference_keynum
            else:
                pc_processed = pci.rk_and_fo
        harmonics = self._harmonics(ratios)
        for h, p in zip(harmonics, pc_processed):
            output.append([h] + p)
        return(output)

    def _intervals_and_resultant_tones(self):
        data = []
        for i, y in enumerate([lambda x: tuple(x), lambda x: x[0] / x[1], lambda x: x[0] - x[1], lambda x: x[0] + x[1]]):
            raw_pitch_info = [ y(sorted(x, reverse=True)) for x in list(combinations(self.ratios[::-1], 2)) ]
            data.append(sorted(list(dict.fromkeys(raw_pitch_info))))
            intervals = []
            if i == 1:
                pitch_pairs = data[0]
                intervals = data[1]
            elif i > 1:
                pitch_pairs = list(product(self.ratios, list(dict.fromkeys(raw_pitch_info))))
                intervals = [x[0] / x[1] for x in pitch_pairs]
            tuneable_intervals = []
            tuneable_resultant_tones = []
            tuneable_pitch_pairs = []
            non_tuneable_intervals = []
            non_tuneable_resultant_tones = []
            non_tuneable_pitch_pairs = []
            for y in intervals:
                if self._is_tuneable(y):
                    tuneable_intervals.append(y)
                    tpp = []
                    for pp in pitch_pairs:
                        if pp[0] / pp[1] == y:
                            if i == 1:
                                tpp.append(pp)
                            else:
                                tuneable_resultant_tones.append(pp[1])
                                tuneable_resultant_tones = sorted(list(dict.fromkeys(tuneable_resultant_tones)))
                            if i == 1:
                                tuneable_pitch_pairs.append(tpp)
                else:
                    non_tuneable_intervals.append(y)
                    ntpp = []
                    for pp in pitch_pairs:
                        if pp[0] / pp[1] == y:
                            if i == 1:
                                ntpp.append(pp)
                                non_tuneable_resultant_tones.append(pp[1])
                            else:
                                non_tuneable_resultant_tones.append(pp[1])
                                non_tuneable_resultant_tones = sorted(list(dict.fromkeys(non_tuneable_resultant_tones)))
                            if i == 1:
                                non_tuneable_pitch_pairs.append(ntpp)
            if i == 1:
                data.append([tuneable_intervals, tuneable_pitch_pairs, non_tuneable_intervals, non_tuneable_pitch_pairs])
            elif i > 1:
                for x in non_tuneable_resultant_tones:
                    if x in tuneable_resultant_tones:
                        non_tuneable_resultant_tones.remove(x)
                for tct in tuneable_resultant_tones:
                    tpp = []
                    for pp in pitch_pairs:
                        if tct == pp[1] and self._is_tuneable(pp[0] / pp[1]):
                            tpp.append(pp)
                    tuneable_pitch_pairs.append(tpp)
                for ntct in non_tuneable_resultant_tones:
                    ntpp = []
                    for pp in pitch_pairs:
                        if ntct == pp [1] and self._is_tuneable(pp[0] / pp[1]) == False:
                            ntpp.append(pp)
                    non_tuneable_pitch_pairs.append(ntpp)
                data.append([tuneable_resultant_tones, tuneable_pitch_pairs, non_tuneable_resultant_tones, non_tuneable_pitch_pairs])
        return(data[1:])

    def _intervals_sequential(self):
        intervals_sequential = [ self.ratios[n] / self.ratios[n - 1] for n in range(1,len(self.ratios)) ]
        return(intervals_sequential)

    def _inversion(self):
        ratios = self.ratios
        inverted_intervals = [ ratios[n - 1] / ratios[n] for n in range(1,len(ratios))]
        inversion = [ratios[-1]]
        for x in inverted_intervals:
            inversion.append(inversion[-1] * x)
        min_inversion = min(inversion)
        correction = fractions.Fraction(min(ratios), min(inversion))
        transposed_inversion = [x * correction for x in inversion]
        return(sorted(transposed_inversion))

    def _is_tuneable(self, ratio):
        is_tuneable = False
        if ratio in self.tuneable_intervals:
            is_tuneable = True
        return(is_tuneable) 

    def _least_common_partial(self):
        least_common_partial = utilities_general.lcm(self.harmonics)
        least_common_partial_freq = least_common_partial * self.periodicity_pitch
        return([least_common_partial, least_common_partial_freq])

    def _pc_plus_resultant_tones(self):
        composite_pc = list(dict.fromkeys(sorted(self.ratios + self.difference_tones + self.summation_tones)))
        return(composite_pc)

    def _periodicity_pitch(self):
        lowest_harmonic = min(self.harmonics)
        lowest_freq = min(self.freqs)
        periodicity_pitch = lowest_freq / lowest_harmonic
        return(periodicity_pitch)


import math
from fractions import Fraction
from functools import reduce

def tuple_to_fraction(t):
    f = Fraction(t[0], t[1])
    return(f)

def tuples_to_fractions(l):
    output = [tuple_to_fraction(x) for x in l]
    return(output)

def flop(lists):
    final = []
    lens = []
    for l in lists:
        lens.append(len(l))
    leng = min(lens)
    for i in range(leng):
        k = []
        for l in lists:
          k.append(l[i])
        final.append(k)
    return(final)

def lcm(list_of_integers):
    output = reduce(lambda x, y: x * y // math.gcd(x, y), list_of_integers)
    return(output)

def convert_data_to_readable_string(d, precision = 5, prefix = False, suffix = False):
    if isinstance(d, float):
        formatted_string = str(round(d, precision))
    else:
        formatted_string = str(d)
    if prefix:
        formatted_string = prefix + formatted_string
    if suffix:
        formatted_string = formatted_string + suffix
    return(formatted_string)


from itertools import compress

#adapted from: https://stackoverflow.com/questions/31843844/how-to-calculate-the-exponents-of-prime-factors-for-a-given-number
#used to create list of primes, and to factor numbers into primes with exponents
class PrimeList():

    def __init__(self, max_val):
        self.primes = self._init_primes(max_val)
        self.extend(len(self.primes))

    def check(self,n):
        n = int(n)
        limit = int(math.sqrt(n))
        return all((n%p for p in self.primes if p<=limit))

    def extend(self,n):
        n = int(n)
        nextnum = self.primes[-1]
        while(self.primes[-1]<n):
            nextnum += 2
            if self.check(nextnum):
                self.primes.append(nextnum)

    def factors(self, n):
        n = int(n)
        x = n
        fact = dict()
        for p in self.primes:
            if p>x:
                break
            while x%p==0:
                x = x/p
                fact[p]=fact.get(p,0)+1
        if x>1:
            e = x if x!=n else int(math.sqrt(n))
            self.extend(e)
            return self.factors(n)
        return sorted(fact.items())

    def is_prime(self, n):
        n = int(n)
        self.extend(int(math.sqrt(n)))
        return self.check(n)

        #taken from https://stackoverflow.com/questions/2068372/fastest-way-to-list-all-primes-below-n/3035188#3035188
    def _init_primes(self, max_val):
        """ Returns a list of primes < n for n > 2 """
        sieve = bytearray([True]) * (max_val//2+1)
        for i in range(1,int(max_val**0.5)//2+1):
            if sieve[i]:
                sieve[2*i*(i+1)::2*i+1] = bytearray((max_val//2-2*i*(i+1))//(2*i+1)+1)
        return [2,*compress(range(3,max_val,2), sieve[1:])]


import ast
import csv
import fractions
import math
import os
from . import utilities_general, utilities_music, prime_list, constants

LONG_LIST_OF_PRIMES = prime_list.PrimeList(2**24) #if this value is too low there will be errors when trying to factor very large numbers
DEFAULT_MONZO_PRIMES = prime_list.PrimeList(48).primes #all primes up to 47 by default

class Pitch():

    def __init__(
        self,
        p = (1, 1),
        rp = "A4",
        rf = 440.0,
        precision = 5):
        self.reference_pitch = rp
        self.reference_freq = rf
        self.precision = precision
        self._vector_primes = DEFAULT_MONZO_PRIMES
        self.rk_and_fo = self._reference_keynum_and_fund_offset_from_pitch_letter_name_and_octave()
        self.reference_keynum = self.rk_and_fo[0]
        self._fund_offset = self.rk_and_fo[1]
        if isinstance(p, fractions.Fraction):
            self.ratio = p
            self.monzo = self._monzo_from_ratio()
        elif isinstance(p, list):
            self.monzo = self._trim_monzo(p)
            self.ratio = self._ratio_from_monzo()
        elif isinstance(p, tuple):
            self.ratio = utilities_general.tuple_to_fraction(p)
            self.monzo = self._monzo_from_ratio()
        self.constituent_primes = self._constituent_primes()
        self.freq = self._freq()
        self.keynum = self._keynum()
        self.keynum_class = self.keynum % 12
        self.distance_in_cents_from_reference = self._distance_in_cents_from_reference()
        self.notation = self._notation()
        self.accidental_string = self.notation[0]
        self.letter_name = self.notation[1]
        self.letter_name_and_octave_and_cents = self._letter_name_and_octave_and_cents()
        self.complement = self._complement()
        self.normalized_ratio = self._normalized_ratio(self.ratio)
        self.normalized_monzo = self._normalized_monzo(self.monzo)
        self.normalized_complement = self._normalized_ratio(self.complement)
        if self.accidental_string != "undefined":
            self.num_symbols = len(self.accidental_string)
        else:
            self.num_symbols = "undefined"
        self.harmonic_distance = self._harmonic_distance(self.monzo)
        self.normalized_harmonic_distance = self._harmonic_distance(self.normalized_monzo)
        self.pitch_info = self._pitch_info()

    def create_strings_for_print_and_txt(self, variety = "basic"):
        basic_info_strings = [
            "",
            "BASIC INFO",
            "ratio: " + str(self.ratio),
            "monzo: " + str(self.monzo),
            "constituent primes: " + str(self.constituent_primes),
            "frequency (Hz): " + utilities_general.convert_data_to_readable_string(self.freq, precision = self.precision),
            "MIDI key number: " + utilities_general.convert_data_to_readable_string(self.keynum, precision = self.precision),
            "distance from 1/1 (cents): " + utilities_general.convert_data_to_readable_string(self.distance_in_cents_from_reference, precision = self.precision),
            "harmonic distance: " + utilities_general.convert_data_to_readable_string(self.harmonic_distance, precision = self.precision),
            "Helmholtz-Ellis notation (text string, letter name): " + str((self.accidental_string, self.letter_name)),
            "12-ED2 pitch and cent deviation: " + str(self.letter_name_and_octave_and_cents),
            ""]
        normalized_info_strings = [
            "",
            "NORMALIZED INFO",
            "normalized ratio: " + str(self.normalized_ratio),
            "normalized monzo: " + str(self.normalized_monzo),
            "MIDI key number class: " + utilities_general.convert_data_to_readable_string(self.keynum_class, precision = self.precision),
            "normalized complement: " + str(self.normalized_complement),
            "normalized harmonic distance: " + utilities_general.convert_data_to_readable_string(self.normalized_harmonic_distance, precision = self.precision),
            ""]
        reference_info_strings = [
            "",
            "REFERENCE INFO",
            "reference pitch (1/1): " + str(self.reference_pitch),
            "reference key number: " + utilities_general.convert_data_to_readable_string(self.reference_keynum, precision = self.precision),
            "reference frequency: " + utilities_general.convert_data_to_readable_string(self.reference_freq, precision = self.precision, suffix = " Hz"),
            ""]
        if variety == "basic":
            output = basic_info_strings
        elif variety == "normalized":
            output = normalized_info_strings
        elif variety == "reference":
            output = reference_info_strings
        elif variety == "all":
            output = basic_info_strings + normalized_info_strings[1:] + reference_info_strings[1:]
        return(output)

    def get_enharmonics(
        self, 
        tolerance = 1.95, 
        limit = 23, 
        exclude_primes = [], 
        max_symbols = 2, 
        max_hd = 30, 
        max_candidates = 10, 
        sort_by = "tolerance"):

        possible_enharmonics = []
        fund_offset = self._fund_offset
        vector_primes = self._vector_primes
        reference_pc_height = Pitch(p = self.normalized_monzo, rp = self.reference_pitch, rf = self.reference_freq).distance_in_cents_from_reference % 1200.0
        path_to_lookup_table = constants.RESOURCES_DIRECTORY + "/enharmonic_lookup_table.csv"
        with open(path_to_lookup_table, newline="") as f:
            reader = csv.reader(f)
            data = list(reader)
        min_candidate_pc_height = (reference_pc_height - tolerance) % 1200.0
        max_candidate_pc_height = (reference_pc_height + tolerance) % 1200.0
        starting_index_orientation = False
        if min_candidate_pc_height > max_candidate_pc_height:
            index_of_candidate = 0
            starting_index_orientation = "within tolerance"
        else:
            index_of_candidate = round((min_candidate_pc_height / 1200.0) * len(data))
        candidate_data = data[index_of_candidate]
        candidate_pc_height = ast.literal_eval(candidate_data[1])
        if not starting_index_orientation:
            if candidate_pc_height <= reference_pc_height:
                starting_index_orientation = "ascend only"
            elif candidate_pc_height >= reference_pc_height:
                starting_index_orientation = "descend only"
            else:
                starting_index_orientation = "within tolerance"
                        
        def get_enharmonics_from_span(orientation):
            nonlocal candidate_pc_height
            nonlocal index_of_candidate
            nonlocal min_candidate_pc_height
            nonlocal max_candidate_pc_height
                    
            if orientation == "ascend":
                def while_condition():
                    nonlocal candidate_pc_height
                    nonlocal max_candidate_pc_height
                    return(candidate_pc_height <= max_candidate_pc_height)
                incr_value = 1
                def break_condition():
                    nonlocal candidate_pc_height
                    nonlocal max_candidate_pc_height
                    return(candidate_pc_height > max_candidate_pc_height)
            else:
                def while_condition():
                    nonlocal candidate_pc_height
                    nonlocal min_candidate_pc_height
                    return(candidate_pc_height >= min_candidate_pc_height)
                incr_value = -1
                def break_condition():
                    nonlocal candidate_pc_height
                    nonlocal min_candidate_pc_height
                    return(candidate_pc_height < min_candidate_pc_height)
                        
            if max_candidate_pc_height < min_candidate_pc_height:
                def range_condition():
                    nonlocal candidate_pc_height
                    nonlocal min_candidate_pc_height
                    nonlocal max_candidate_pc_height
                    return( ((candidate_pc_height - 1200.0) <= max_candidate_pc_height and candidate_pc_height >= min_candidate_pc_height) or (candidate_pc_height <= max_candidate_pc_height and (candidate_pc_height + 1200.0) >= min_candidate_pc_height))
            else:
                def range_condition():
                    nonlocal candidate_pc_height
                    nonlocal min_candidate_pc_height
                    nonlocal max_candidate_pc_height
                    return(candidate_pc_height <= max_candidate_pc_height and candidate_pc_height >= min_candidate_pc_height)
                    
            while while_condition():
                candidate_data = data[index_of_candidate]
                candidate_pc_height = ast.literal_eval(candidate_data[1])
                if range_condition():
                    limit_ok = True
                    candidate_monzo = ast.literal_eval(candidate_data[0])
                    max_prime_index = vector_primes.index(limit)
                    for i in range(max_prime_index + 1, len(candidate_monzo)):
                        if candidate_monzo[i] != 0:
                            limit_ok = False
                    forbidden_prime_indices = [vector_primes.index(i) for i in exclude_primes]
                    for i in forbidden_prime_indices:
                        if candidate_monzo[i] != 0:
                            limit_ok = False
                    if limit_ok:
                        if abs(candidate_pc_height - reference_pc_height) > tolerance:
                            if candidate_pc_height > reference_pc_height:
                                candidate_monzo[0] = candidate_monzo[0] - 1
                            else:
                                candidate_monzo[0] = candidate_monzo[0] + 1
                        candidate_original_octave_difference = round((self.distance_in_cents_from_reference - reference_pc_height) / 1200)
                        candidate_monzo[0] = candidate_monzo[0] + candidate_original_octave_difference
                        candidate_pci = Pitch(p = candidate_monzo, rp = self.reference_pitch, rf = self.reference_freq)
                        if candidate_pci.ratio != self.ratio:
                            candidate_hd = candidate_pci.harmonic_distance
                            if candidate_hd <= max_hd:
                                candidate_num_symbols = candidate_pci.num_symbols
                                enharmonic_difference = round(candidate_pci.distance_in_cents_from_reference - self.distance_in_cents_from_reference, self.precision)
                                if candidate_num_symbols <= max_symbols:
                                    formatted_candidate = [candidate_pci.ratio, enharmonic_difference, candidate_hd, abs(enharmonic_difference)]
                                    possible_enharmonics.append(formatted_candidate)
                index_of_candidate = index_of_candidate + incr_value
                if break_condition():                                               
                    break

        if starting_index_orientation == "ascend only":
            get_enharmonics_from_span("ascend")
        elif starting_index_orientation == "descend only":
            get_enharmonics_from_span("descend")
        if starting_index_orientation == "within tolerance":
            ascending_index = index_of_candidate + 0
            index_of_candidate -= 1
            if index_of_candidate == 0:
                candidate_pc_height = 1200.0
            else:
                candidate_data = data[index_of_candidate]
                candidate_pc_height = ast.literal_eval(candidate_data[1])
            get_enharmonics_from_span("descend")
            index_of_candidate = ascending_index
            candidate_data = data[index_of_candidate]
            candidate_pc_height = ast.literal_eval(candidate_data[1])
            get_enharmonics_from_span("ascend")
        if len(possible_enharmonics) > 0:
            if sort_by == "harmonic distance":
                possible_enharmonics.sort(key=lambda x: x[3])
                possible_enharmonics.sort(key=lambda x: x[2])
            else:
                possible_enharmonics.sort(key=lambda x: x[2])
                possible_enharmonics.sort(key=lambda x: x[3])
        possible_enharmonics = possible_enharmonics[:max_candidates]
        if len(possible_enharmonics) > 0:
            for x in possible_enharmonics:
                x[3] =  self.ratio / x[0]
        return(possible_enharmonics)
        
    def print_info(self, variety = "basic"):
        strings_to_print = self.create_strings_for_print_and_txt(variety = variety)
        for x in strings_to_print:
            print(x)

    def print_enharmonics_info(
        self, 
        tolerance = 1.95, 
        limit = 23, 
        exclude_primes = [], 
        max_symbols = 2, 
        max_hd = 30, 
        max_candidates = 10, 
        sort_by = "tolerance"):
        enharmonics_info = self.get_enharmonics(
            tolerance = tolerance, 
            limit = limit, 
            exclude_primes = exclude_primes, 
            max_symbols = max_symbols, 
            max_hd = max_hd, 
            max_candidates = max_candidates, 
            sort_by = sort_by)
        num_enharmonics = len(enharmonics_info)
        header_strings = self._create_strings_for_enharmonics_header(
                tolerance = tolerance,
                limit = limit,
                exclude_primes = exclude_primes,
                max_symbols = max_symbols,
                max_hd = max_hd,
                max_candidates = max_candidates,
                sort_by = sort_by,
                num_qualified_candidates = num_enharmonics)
        print("")
        for x in header_strings:
            print(x)
        print("")
        if num_enharmonics > 0:
            enharmonics_strings = self._create_strings_for_enharmonics_info(enharmonics_info)
            for x in enharmonics_strings:
                print(x)

    def transpose(self, interval):
        new_ratio = self.ratio * Pitch(p = interval).ratio 
        self.update(p = new_ratio)

    def update(self, p = False, rp = False, rf = False, precision = False):
        if not p:
            p = self.monzo
        if not rp:
            rp = self.reference_pitch
        if not rf:
            rf = self.reference_freq
        if not precision:
            precision = self.precision
        self.__init__(p = p, rp = rp, rf = rf, precision = precision)

    def write_enharmonics_info_to_csv(
        self, 
        tolerance = 1.95, 
        limit = 23, 
        exclude_primes = [], 
        max_symbols = 2, 
        max_hd = 30, 
        max_candidates = 10, 
        sort_by = "tolerance",
        output_directory = False,
        filename = "enharmonic_candidates.csv"):
        enharmonics_info = self.get_enharmonics(
            tolerance = tolerance, 
            limit = limit, 
            exclude_primes = exclude_primes, 
            max_symbols = max_symbols, 
            max_hd = max_hd, 
            max_candidates = max_candidates, 
            sort_by = sort_by)
        num_enharmonics = len(enharmonics_info)
        header_strings = self._create_strings_for_enharmonics_header(
            tolerance = tolerance,
            limit = limit,
            exclude_primes = exclude_primes,
            max_symbols = max_symbols,
            max_hd = max_hd,
            max_candidates = max_candidates,
            sort_by = sort_by,
            num_qualified_candidates = num_enharmonics)
        formatted_header = []
        for x in header_strings:
            formatted_header.append([" ", x])
        formatted_header.append(" ")
        final_info = formatted_header
        if output_directory == False:
            output_directory = os.getcwd()
        path_to_write_file = output_directory + "/" + filename
        if num_enharmonics > 0:
            data_types = [
                "", 
                "ratio", 
                "monzo", 
                "freq (Hz)", 
                "keynum", 
                "primes", 
                "hd", 
                "HEJI", 
                "12-ED2", 
                "mel. ratio", 
                "enh. size (cents)"]
            count = 0
            processed_info = []
            for x in enharmonics_info:
                count += 1
                ratio = x[0]
                delta = x[1]
                ei = x[3]
                enharmonic_ici = Pitch(
                    p = (ratio.numerator, ratio.denominator), 
                    rp = self.reference_pitch, 
                    rf = self.reference_freq, 
                    precision = self.precision)
                stats_to_be_added = [count,
                    "'" + str(enharmonic_ici.ratio) + "'",
                    str(enharmonic_ici.monzo),
                    utilities_general.convert_data_to_readable_string(enharmonic_ici.freq, precision = self.precision),
                    utilities_general.convert_data_to_readable_string(enharmonic_ici.keynum, precision = self.precision),
                    str(enharmonic_ici.constituent_primes),
                    utilities_general.convert_data_to_readable_string(enharmonic_ici.harmonic_distance, precision = self.precision),
                    str(enharmonic_ici.notation),
                    "'" + str(enharmonic_ici.letter_name_and_octave_and_cents) + "'",
                    str(self._fraction_to_proportional_ratio_string(ei)),
                    str(delta)]
                processed_info.append(stats_to_be_added)
            final_info = formatted_header + [data_types] + processed_info
            with open(path_to_write_file, "w") as output:
                writer = csv.writer(output, lineterminator='\n')
                writer.writerows(final_info)
        print("file written to " + path_to_write_file)

    def write_enharmonics_info_to_txt(
        self, 
        tolerance = 1.95, 
        limit = 23, 
        exclude_primes = [], 
        max_symbols = 2, 
        max_hd = 30, 
        max_candidates = 10, 
        sort_by = "tolerance",
        output_directory = False,
        filename = "enharmonic_candidates.txt"):
        enharmonics_info = self.get_enharmonics(
            tolerance = tolerance, 
            limit = limit, 
            exclude_primes = exclude_primes, 
            max_symbols = max_symbols, 
            max_hd = max_hd, 
            max_candidates = max_candidates, 
            sort_by = sort_by)
        num_enharmonics = len(enharmonics_info)
        header_strings = self._create_strings_for_enharmonics_header(
            tolerance = tolerance,
            limit = limit,
            exclude_primes = exclude_primes,
            max_symbols = max_symbols,
            max_hd = max_hd,
            max_candidates = max_candidates,
            sort_by = sort_by,
            num_qualified_candidates = num_enharmonics)
        if output_directory == False:
            output_directory = os.getcwd()
        path_to_write_file = output_directory + "/" + filename
        with open(path_to_write_file, "w") as output:
            for x in header_strings:
                output.write(x + "\n")
            output.write(" \n")
            if num_enharmonics > 0:
                enharmonics_strings = self._create_strings_for_enharmonics_info(enharmonics_info)
                for x in enharmonics_strings:
                    output.write(x + "\n")
        print("file written to " + path_to_write_file)

    def write_info_to_txt(self, variety = "all", output_directory = False, filename = "pitch_info.txt"):
        strings_to_write = self.create_strings_for_print_and_txt(variety = variety)
        if output_directory == False:
            output_directory = os.getcwd()
        path_to_write_file = output_directory + "/" + filename
        with open(path_to_write_file, "w") as output:
            for x in strings_to_write:
                output.write(x + "\n")
        print("file written to " + path_to_write_file)

    def _complement(self):
        complement = 2 / self.ratio
        return(complement)

    def _constituent_primes(self):
        monzo = self.monzo
        vector_primes = self._vector_primes
        vector_primes = self._lengthen_vector_primes(monzo, vector_primes)
        constituent_primes = []
        for i, x in enumerate(monzo):
            if abs(x) > 0:
                constituent_primes.append(vector_primes[i])
        return(constituent_primes)

    def _create_strings_for_enharmonics_header(
        self, 
        tolerance = 1.95, 
        limit = 23, 
        exclude_primes = [], 
        max_symbols = 2, 
        max_hd = 30, 
        max_candidates = 10, 
        sort_by = "tolerance",
        num_qualified_candidates = 0):

        def excluded_primes_string():
            output = []
            if len(exclude_primes) > 0:
                output.append("excluded primes: " + str(exclude_primes))
            return(output)
                    
        header_strings = ["ORIGINAL PITCH INFO"] + self.create_strings_for_print_and_txt(variety = "basic")[2:] + [
            "ENHARMONIC SELECTION CRITERIA", 
            "tolerance (cents): " + utilities_general.convert_data_to_readable_string(tolerance, precision = 5),
            "prime limit: " + str(limit) ] + excluded_primes_string() + ["maximum number of HEJI symbols: " + str(max_symbols),
            "maximum mumber of candidates: " + str(max_candidates),
            "sorted by: " + str(sort_by),
            "total number of qualifying candidates: " + str(num_qualified_candidates)]
                    
        return(header_strings)

    def _create_strings_for_enharmonics_info(self, enharmonics_info):
        enharmonics_info_strings = []
        for i, x in enumerate(enharmonics_info):
            count = i + 1
            ratio = x[0]
            delta = x[1]
            if delta > 0:
                delta = "+" + str(delta)
            else:
                delta = str(delta)
            ei = x[3]
            enharmonics_info_strings.append("ENHARMONIC NO. " + str(count))
            enharmonic_ici = Pitch(p = (ratio.numerator, ratio.denominator), rp = self.reference_pitch, rf = self.reference_freq, precision = self.precision)
            enharmonic_info_strings = enharmonic_ici.create_strings_for_print_and_txt(variety = "basic")[2:][:-1]
            for x in enharmonic_info_strings:
                enharmonics_info_strings.append(x)
            enharmonics_info_strings.append("melodic interval from " + str(self.ratio) + ": " + self._fraction_to_proportional_ratio_string(ei))
            enharmonics_info_strings.append("enharmonic interval size (cents): " + str(delta))
            enharmonics_info_strings.append("")
        return(enharmonics_info_strings)

    def _distance_in_cents_from_reference(self):
        distance_in_cents_from_reference = (self.keynum - self.reference_keynum) * 100
        return(distance_in_cents_from_reference)

    def _fraction_to_proportional_ratio_string(self, f):
        return(str(f.numerator) + ":" + str(f.denominator))

    def _freq(self):
        freq = float(self.reference_freq * self.ratio)
        return(freq)

    def _harmonic_distance(self, monzo):
        vector_primes = self._vector_primes
        vector_primes = self._lengthen_vector_primes(monzo, vector_primes)
        harmonic_distance = 0
        for i in range(0, len(monzo)):
            current_prime = vector_primes[i]
            current_exponent = abs(monzo[i])
            current_hd = current_exponent * math.log2(current_prime)
            harmonic_distance += current_hd
        return(harmonic_distance)

    def _keynum(self):
        keynum = utilities_music.cpsmidi(self.freq, 
            ref_freq = self.reference_freq, 
            ref_keynum = self.reference_keynum)
        return(keynum) 

    def _letter_name_and_octave_and_cents(self):
        pitch_class_letter_name_index = math.floor(self.keynum_class)
        pitch_octave = int(divmod(self.keynum, 12)[0] - 1)
        pitch_cents = self.keynum_class % 1
        if (1 - pitch_cents) < pitch_cents:
            pitch_cents = 1 - pitch_cents
            cents_sign = "-"
            pitch_class_letter_name_index += 1
        else:
            cents_sign = "+"
        pitch_class_letter_name_index = int(pitch_class_letter_name_index) % 12
        pitch_class_letter_name_strings = ["C", "C#/Db", "D", "D#/Eb", "E", "F", "F#/Gb", "G", "G#/Ab", "A", "A#/Bb", "B"]
        pitch_class_letter_name = pitch_class_letter_name_strings[pitch_class_letter_name_index]
        return(pitch_class_letter_name + str(pitch_octave) + " " + cents_sign + str(round(pitch_cents * 100, self.precision)))

    def _lengthen_vector_primes(self, monzo, vector_primes):
        if len(monzo) > len(vector_primes):
            x = vector_primes[-1:][0]
            while len(vector_primes) < len(monzo):
                vector_primes = prime_list.PrimeList(max_val = x + 1).primes
                x += 1
                if len(vector_primes) >= len(monzo):
                    break
        return(vector_primes)

    def _monzo_from_ratio(self):
        if self.ratio == fractions.Fraction(1, 1):
            monzo = [0] * len(self._vector_primes)
        else:              
            numerators = LONG_LIST_OF_PRIMES.factors(self.ratio.numerator)
            denominators = LONG_LIST_OF_PRIMES.factors(self.ratio.denominator)
            max_default_prime = self._vector_primes[-1:][0]
            all_primes = []
            for x in numerators + denominators:
                all_primes.append(x[0])
            max_prime_in_ratio = max(all_primes)
            if max_prime_in_ratio > max_default_prime:
                self._vector_primes = prime_list.PrimeList(max_val = max_prime_in_ratio + 1).primes
            monzo = []
            for i in range(0, len(self._vector_primes)):
                monzo.append(0)
            for x in numerators:
                for i in range(0, len(self._vector_primes)):
                    current_prime = self._vector_primes[i]
                    if x[0] == current_prime:
                        monzo[i] += x[1]
            for x in denominators:
                for i in range(0, len(self._vector_primes)):
                    current_prime = self._vector_primes[i]
                    if x[0] == current_prime:
                        monzo[i] -= x[1]
        trimmed_monzo = self._trim_monzo(monzo)
        return(trimmed_monzo)

    def _normalized_monzo(self, monzo):
        ratio = self.ratio
        normalized_monzo = [] + monzo
        while (ratio.numerator - ratio.denominator) < 0:
            normalized_monzo[0] += 1
            ratio = Pitch(p = normalized_monzo).ratio
            if (ratio.numerator - ratio.denominator) >= 0:
                break
        while (ratio.numerator - ratio.denominator) >= ratio.denominator:
            normalized_monzo[0] -= 1
            ratio = Pitch(p = normalized_monzo).ratio
            if (ratio.numerator - ratio.denominator) < ratio.denominator:
                break                
        return(normalized_monzo)

    def _normalized_ratio(self, ratio):
        num = ratio.numerator
        den = ratio.denominator
        condition = num / den < 1
        while condition:
            num = num * 2
            condition = num / den < 1
        condition = num / den >= 2
        while condition:
            den = den * 2
            condition = num / den >= 2
        normalized_ratio = fractions.Fraction(num, den)
        return(normalized_ratio)
        
    def _normalized_harmonic_distance(self):
        normalized_harmonic_distance = Pitch(p = self.normalized_monzo)._harmonic_distance()
        return(normalized_harmonic_distance)

    def _notation(self):
        fund_offset = self._fund_offset
        monzo = self.monzo
        vector_primes = self._vector_primes
        vector_primes = self._lengthen_vector_primes(monzo, vector_primes)
        accidental_undefined = False
        letter_name_undefined = False
        net_3 = 0 + fund_offset # keep track of how many P5s we are moving at all times
        syntonic_arrows = 0
        accidental_characters = []
        for i in range(0, len(monzo)):
            current_prime = vector_primes[i]
            current_exp = monzo[i]
            if current_prime == 3:
                net_3 += current_exp
            elif current_prime == 5:
                net_3 += 4 * current_exp
                syntonic_arrows += current_exp
                if abs(syntonic_arrows) > 4:
                    accidental_undefined = True
            elif current_prime == 7:
                net_3 -= 2 * current_exp
                double_seven, single_seven = divmod(abs(current_exp), 2)
                if current_exp > 0:
                    for _ in range(single_seven):
                        accidental_characters.append("<")
                    for _ in range(double_seven):
                        accidental_characters.append(",")
                elif current_exp < 0:
                    for _ in range(single_seven):
                        accidental_characters.append(">")
                    for _ in range(double_seven):
                        accidental_characters.append(".")
            elif current_prime > 47:
                accidental_undefined = True
                letter_name_undefined = True
            remaining_prime_adjustments = [
                [11, -1, ["4", "5"]],
                [13, 3, ["0", "9"]],
                [17, 7, [":", ";"]],
                [19, -3, ["/", "*"]],
                [23, 6, ["3", "6"]],
                [29, -2, ["2", "7"]],
                [31, 0, ["1", "8"]],
                [37, 2, ["á", "à"]],
                [41, 4, ["+", "-"]],
                [43, -1, ["é", "è"]],
                [47, 6, ["í", "ì"]]]
            for x in remaining_prime_adjustments:
                if current_prime == x[0]:
                    exp_adj = x[1]
                    pos_str = x[2][0]
                    neg_str = x[2][1]
                    net_3 = net_3 + (exp_adj * current_exp)
                    if current_exp > 0:
                        for _ in range(current_exp):
                            accidental_characters.append(pos_str)
                    elif current_exp < 0:
                        for _ in range(abs(current_exp)):
                            accidental_characters.append(neg_str)

        def add_5_limit_sign(acc_chars):
            for i, acc_char in enumerate(acc_chars):
                if syntonic_arrows == i - 4:
                    if abs(net_3) < 4:
                        if syntonic_arrows == 0 and len(accidental_characters) == 0:
                            accidental_characters.insert(0, acc_char)
                        elif syntonic_arrows != 0:
                            accidental_characters.insert(0, acc_char)
                    else:
                        accidental_characters.insert(0, acc_char)
            return()

        if abs(net_3) < 4:
            add_5_limit_sign(["N", "q", "p", "o", "n", "m", "l", "k", "M"])
        elif abs(net_3) > 3 and abs(net_3) < 11:
            if net_3 > 0:
                add_5_limit_sign(["P", "y", "x", "w", "v", "u", "t", "s", "O"])
            else:
                add_5_limit_sign(["L", "h", "g", "f", "e", "d", "c", "b", "K"])
        elif abs(net_3) > 10:
            if net_3 > 0:
                add_5_limit_sign(["Q", "Y", "X", "W", "V", "U", "T", "S", "R"])
            else:
                add_5_limit_sign(["J", "H", "G", "F", "E", "D", "C", "B", "I"])
        excess_3_beyond_doublesharps_and_flats = abs(net_3) - 17
        if excess_3_beyond_doublesharps_and_flats > 0:
            num_excess_signs = math.ceil(excess_3_beyond_doublesharps_and_flats / 7)
            double_three, single_three = divmod(num_excess_signs, 2)
            if net_3 > 0:
                for _ in range(single_three):
                    accidental_characters.insert(1, "v")
                for _ in range(double_three):
                    accidental_characters.insert(1, "V")
            elif net_3 < 0:
                for _ in range(single_three):
                    accidental_characters.insert(1, "e")
                for _ in range(double_three):
                    accidental_characters.insert(1, "E")
        accidental_characters = list(reversed(accidental_characters))
        accidental_string = ""
        letter_names = ["D", "A", "E", "B", "F", "C", "G"]
        if letter_name_undefined:
            letter_name = "undefined"
        else:
            letter_name = letter_names[net_3 % 7]
        if accidental_undefined:
            accidental_string = "undefined"
        else:
            for x in accidental_characters:
                accidental_string += x
        return((accidental_string, letter_name))

    def _pitch_info(self):
        basic_info = [
            self.ratio,
            self.monzo,
            self.constituent_primes,
            self.freq,
            self.keynum,
            self.distance_in_cents_from_reference,
            self.harmonic_distance,
            (self.accidental_string, self.letter_name),
            self.letter_name_and_octave_and_cents]
        normalized_info = [
            self.normalized_ratio,
            self.normalized_monzo,
            self.keynum_class,
            self.normalized_complement,
            self.normalized_harmonic_distance]
        reference_info = [
            self.reference_pitch,
            self.reference_keynum,
            self.reference_freq]
        return([basic_info, normalized_info, reference_info])

    def _ratio_from_monzo(self):
        monzo = self.monzo
        vector_primes = self._vector_primes
        vector_primes = self._lengthen_vector_primes(monzo, vector_primes)
        numerator = 1
        denominator = 1
        for i in range(0, len(monzo)):
            current_prime = vector_primes[i]
            current_exp = monzo[i]
            if current_exp > 0:
                numerator = numerator * (current_prime ** current_exp)
            if current_exp < 0:
                denominator = denominator * (current_prime ** abs(current_exp))
        ratio = utilities_general.tuple_to_fraction((numerator, denominator))
        return(ratio)

    def _reference_keynum_and_fund_offset_from_pitch_letter_name_and_octave(self):
        rp_string = str(self.reference_pitch) + ""
        possible_letter_names = ["F", "C", "G", "D", "A", "E", "B"]
        possible_keynum_classes = [5, 0, 7, 2, 9, 4, 11]
        letter_name = False
        octave_number = False
        output = False
        fund_offset = 0
        if rp_string[0] in possible_letter_names:
            letter_name = rp_string[0]
            keynum_class = possible_keynum_classes[possible_letter_names.index(letter_name)]
            fund_offset = possible_letter_names.index(letter_name) - 3
            rp_string = rp_string[1:]
        if letter_name is not False and len(rp_string) > 0:
            if rp_string[-1].isdigit():
                octave_number = int(rp_string[-1])
                if len(rp_string) > 1 and rp_string[-2] == "-":
                    octave_number = octave_number * -1
                    rp_string = rp_string[:-1]
                rp_string = rp_string[:-1]
                for x in rp_string:
                    if x == "#":
                        fund_offset += 7
                        keynum_class += 1
                    elif x == "x":
                        fund_offset += 14
                        keynum_class += 2
                    elif x == "b":
                        fund_offset -= 7
                        keynum_class -= 1
                keynum = 12 + (12 * octave_number) + keynum_class
                output = [keynum, fund_offset]
        return(output)

    def _trim_monzo(self, monzo):
        index_of_highest_nonzero_exponent = 0
        for i, x in enumerate(monzo):
            if abs(x) > 0:
                index_of_highest_nonzero_exponent = i
        trimmed_monzo = monzo[:index_of_highest_nonzero_exponent + 1]
        return(trimmed_monzo)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="jitools",
    version="0.1",
    license="GNU GPLv3",
    author="M.O. Abbott",
    author_email="moab_bot@protonmail.com",
    description="a Python-based set of utilities for just intonation (JI) pitch and pitch collection research and analysis",
    url="https://github.com/moab-bot/jitools",
    download_url = 'https://github.com/moab-bot/jitools/archive/v_01.tar.gz',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8.5',
)



from math import log2

def cpsmidi(freq, ref_freq = 440, ref_keynum = 69):
    keynum = ref_keynum + 12 * log2(freq / ref_freq)
    return(keynum)

def midicps(keynum, ref_keynum = 69, ref_freq = 440):
    freq = 2**((keynum - ref_keynum) / 12) * ref_freq
    return(freq)